using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Collections;

namespace tsp
{
    public partial class TSPAlgoritam : Form
    {
        private ArrayList brojIter = new ArrayList();
        private ArrayList velPop = new ArrayList();
        private ArrayList vjerMut = new ArrayList();
        private ArrayList statusBest = new ArrayList();
        private ArrayList Populacija = new ArrayList();
        private ArrayList listaNajboljih = new ArrayList();

        
        Random rnd = new Random();
        private double[][] udaljenosti;
        private int brGradova;

        int index_bI = 0;
        int index_vM = 0;
        int index_vP = 0;

        public TSPAlgoritam()
        {
            InitializeComponent();
            ucitajParametre();
            ucitajStatus();
            this.textBoxBrojIter.Text = brojIter[index_bI].ToString();
            this.textBoxVelPop.Text = velPop[index_vP].ToString();
            this.textBoxVjerMut.Text = vjerMut[index_vM].ToString();
            this.listBoxStatistika.Items.Add("Parametri uitani!");
        }

        private void ucitajParametre()
        {
            XmlReader parametri = new XmlTextReader("parametri.xml");

            while (parametri.Read())
            {
                if (parametri.NodeType == XmlNodeType.Element)
                {
                    if (parametri.LocalName.StartsWith("vP"))
                    {
                        velPop.Add(System.Convert.ToInt32(parametri.ReadString()));
                    }
                    if (parametri.LocalName.StartsWith("vM"))
                    {
                        vjerMut.Add(parametri.ReadString());
                    }
                    if (parametri.LocalName.StartsWith("bI"))
                    {
                        brojIter.Add(System.Convert.ToInt32(parametri.ReadString()));
                    }

                }
            }

            parametri.Close();
        }

        private void ucitajStatus()
        {
            int tren_vP = -1;
            int tren_bI = -1;
            string tren_vM = null;
            bool test = true;
            XmlReader status = null;
            try
            {
                int i = 0;
                status = new XmlTextReader("status.xml");
                while (status.Read())
                {
                    switch (status.LocalName)
                    {
                        case "bI":
                            tren_bI = System.Convert.ToInt32(status.ReadString());
                            break;
                        case "vP":
                            tren_vP = System.Convert.ToInt32(status.ReadString());
                            break;
                        case "vM":
                            tren_vM = status.ReadString();
                            break;
                        default:
                            if (status.LocalName.StartsWith("runNo"))
                            {
                                statusBest.Add(status.ReadString());
                                listaNajboljih.Add(System.Convert.ToDouble(statusBest[i++]));
                            }
                            break;
                    }

                }

            }
            catch (Exception ex)
            {
                test = false;
                //MessageBox.Show(this, "GRESKA:"+ex.Message);
            }
            status.Close();
            ////uskladiParametre
            if (test)
            {
                index_bI = brojIter.IndexOf(tren_bI);
                index_vM = vjerMut.IndexOf(tren_vM);
                index_vP = velPop.IndexOf(tren_vP);
            }
        }
        /// <summary>
        /// Glavna metoda koja pokrece i vrti algoritam. Dobiva sve gotovo
        /// (ucitane gradove, pripremljene parametre, provjeren status) i
        /// vrti glavnu petlju algoritma. Troturnirska selekcija, PMX krizanje
        /// i jednostavna mutacija.
        /// </summary>
        private void algoritam()
        {
            for (int i_bI = index_bI; i_bI < brojIter.Count; i_bI++)
            {
                int bI = (int)brojIter[i_bI];
                this.textBoxBrojIter.Text = bI.ToString();
                for (int i_vP = index_vP; i_vP < velPop.Count; i_vP++)
                {
                    int vP = (int)velPop[i_vP];
                    this.textBoxVelPop.Text = vP.ToString();
                    for (int i_vM = index_vM; i_vM < vjerMut.Count; i_vM++)
                    {
                        string vM = (string) vjerMut[i_vM];
                        this.textBoxVjerMut.Text = vM;

                        //ALGORITAM --- prvo ide spremanje statusa
                        XmlDocument doc = new XmlDocument();
                        try
                        {
                            doc.Load("status.xml");

                            doc.LoadXml("<status><bI>" + bI.ToString() + "</bI></status>");
                            XmlElement brojIteracija = doc.CreateElement("vP");
                            brojIteracija.InnerText = vP.ToString();
                            doc.DocumentElement.AppendChild(brojIteracija);

                            XmlElement velPopulacijeXML = doc.CreateElement("vM");
                            velPopulacijeXML.InnerText = vM;
                            doc.DocumentElement.AppendChild(velPopulacijeXML);

                            for (int i_naj = 0; i_naj < statusBest.Count; i_naj++)
                            {
                                XmlElement naj = doc.CreateElement("runNo" + i_naj.ToString());
                                naj.InnerText = (string)statusBest[i_naj];
                                doc.DocumentElement.AppendChild(naj);
                            }
                            doc.Save("status.xml");
                        }
                        catch (Exception)
                        {
                            //stvara novi status file ako stari ne postoji
                            XmlWriter writer = null;
                            // Create an XmlWriterSettings object with the correct options. 
                            XmlWriterSettings settings = new XmlWriterSettings();
                            settings.Indent = true;
                            settings.IndentChars = ("\t");
                            settings.OmitXmlDeclaration = true;

                            // Create the XmlWriter object and write some content.
                            writer = XmlWriter.Create("status.xml", settings);
                            writer.WriteStartDocument();
                            writer.WriteStartElement("status");
                            writer.WriteElementString("bI", bI.ToString());
                            writer.WriteElementString("vP", vP.ToString());
                            writer.WriteElementString("vM", vM);
                            writer.WriteEndElement();
                            writer.WriteEndDocument();
                            writer.Flush();
                            writer.Close();
                        }
                        //Ovo je petlja za 7 ponavljanja jednog seta parametara
                        for (int i_run = statusBest.Count; i_run < 7; i_run++)
                        {
                            this.textBoxRunNo.Text = i_run.ToString();
                            this.Refresh();

                            //OVO JE SAMO ALGORITAM----------

                            //stvori populaciju
                            this.inicijaliziraj(vP);

                            //stvori tocke krizanja
                            int t1 = 0;
                            int t2 = 0;
                            t1 = rnd.Next(brGradova);

                            do
                            {
                                t2 = rnd.Next(brGradova);
                            } while (t1 == t2);
                            //t1 je uvijek manji od t2
                            if (t2 < t1)
                            {
                                int tmp = t2;
                                t2 = t1;
                                t1 = tmp;
                            }
                           
                            //vrti algoritam bI puta
                            for (int i = 0; i < bI; i++)
                            {
                                
                                //odaberi tri jednike
                                int index_jed1 = 0;
                                int index_jed2 = 0;
                                int index_jed3 = 0;
                                index_jed1 = rnd.Next(0, vP);

                                do
                                {
                                    index_jed2 = rnd.Next(0, vP);
                                } while (index_jed1 == index_jed2);

                                do
                                {
                                    index_jed3 = rnd.Next(0, vP);
                                } while ((index_jed1 == index_jed3) || (index_jed2 == index_jed3));

                                Kromosom jedinka1 = (Kromosom)Populacija[index_jed1];
                                Kromosom jedinka2 = (Kromosom)Populacija[index_jed2];
                                Kromosom jedinka3 = (Kromosom)Populacija[index_jed3];
                                
                                //krizaj najbolje dvije, mutiraj dijete i stavi ga u populaciju
                                Kromosom dijete = null;
                                if ((jedinka1.DuljinaPuta >= jedinka2.DuljinaPuta) && (jedinka1.DuljinaPuta >= jedinka3.DuljinaPuta))
                                {
                                    dijete = this.PartiallyMatchedCrossover(jedinka2, jedinka3,t1,t2);
                                    Populacija.Insert(index_jed1,this.jednostavnaMutacija(dijete, vM));
                                    Populacija.Remove(jedinka1);
                                    continue;
                                }

                                if ((jedinka2.DuljinaPuta >= jedinka1.DuljinaPuta) && (jedinka2.DuljinaPuta >= jedinka3.DuljinaPuta))
                                {
                                    dijete = this.PartiallyMatchedCrossover(jedinka1, jedinka3, t1, t2);
                                    Populacija.Insert(index_jed2, this.jednostavnaMutacija(dijete, vM));
                                    Populacija.Remove(jedinka2);
                                    continue;
                                }
                                if ((jedinka3.DuljinaPuta >= jedinka2.DuljinaPuta) && (jedinka3.DuljinaPuta >= jedinka1.DuljinaPuta))
                                {
                                    dijete = this.PartiallyMatchedCrossover(jedinka2, jedinka1, t1, t2);
                                    Populacija.Insert(index_jed3, this.jednostavnaMutacija(dijete, vM));
                                    Populacija.Remove(jedinka3);
                                    continue;
                                }
                               
                            }
                            //nadji najboljeg
                            Kromosom najbolji = this.nadjiNajboljeg();
                            //this.listBoxStatistika.Items.Add(najbolji.ToString());
                            //OVO JE KRAJ ALGORITMA----------

                            listaNajboljih.Add(najbolji.DuljinaPuta);
                            doc.Load("status.xml");
                            XmlElement naj = doc.CreateElement("runNo" + i_run.ToString());
                            naj.InnerText = System.Convert.ToString(najbolji.DuljinaPuta);
                            doc.DocumentElement.AppendChild(naj);

                            doc.Save("status.xml");
                        }
                        

                        //ovo se radi zato jer je u statusBest bilo pohranjeno koliko je runova
                        //izvrseno kad je doslo do prekida, nakon toga mora se uvijek kretati od
                        //0
                        statusBest.Clear();

                        //spremanje rezultata
                        XmlWriter writerRez = null;
                        // Create an XmlWriterSettings object with the correct options. 
                        XmlWriterSettings settingsRez = new XmlWriterSettings();
                        settingsRez.Indent = true;
                        settingsRez.IndentChars = ("\t");
                        settingsRez.OmitXmlDeclaration = true;

                        // Create the XmlWriter object and write some content.
                        writerRez = XmlWriter.Create("rezultat-" + bI + "-" + vP + "-" + vM + ".xml", settingsRez);
                        writerRez.WriteStartDocument();
                        writerRez.WriteStartElement("rezultat");
                        writerRez.WriteElementString("bI", bI.ToString());
                        writerRez.WriteElementString("vP", vP.ToString());
                        writerRez.WriteElementString("vM", vM);
                        int brojac = 0;
                        foreach (double d in listaNajboljih)
                        {
                            writerRez.WriteElementString("najbolji" + brojac.ToString(), System.Convert.ToString(d));
                            brojac++;
                        }
                        listaNajboljih.Clear();
                        writerRez.WriteEndElement();
                        writerRez.WriteEndDocument();
                        writerRez.Flush();
                        writerRez.Close();
                        

                    }
                    index_vM = 0;
                    
                }
                index_vP = 0;
            }
            
        }

        private void buttonUcitaj_Click(object sender, EventArgs e)
        {
            string fileName;
            ArrayList tocke = new ArrayList() ;
            OpenFileDialog dlg = new OpenFileDialog();
            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                fileName = dlg.FileName;

                //lijepi prikaz imena filea
                this.textBoxNazivProblema.Text = fileName.Substring(fileName.LastIndexOf("\\")+1);

                gradovi doc;
                try
                {
                    doc = new gradovi(fileName);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(this, ex.Message);
                    return;
                }

                //uzimanje broja gradova
                brGradova = System.Int32.Parse(doc.GetValue("info", "brojGradova"));
                string[] str = doc.GetValue("tocke", "t0").Split(' ');
                PointF t = new PointF((float)System.Double.Parse(str[0]), (float)System.Double.Parse(str[1]));
                tocke.Add(t);
                PointF min = new PointF(t.X, t.Y);
                PointF max = new PointF(t.X, t.Y);
                //vadjenje koordinata i odredjivanje min i max vrijednosti
                for (int i = 1; i < this.brGradova; i++)
                {
                    str = doc.GetValue("tocke", "t" + i.ToString()).Split(' ');
                    t.X = (float)System.Double.Parse(str[0]);
                    t.Y = (float)System.Double.Parse(str[1]);
                    tocke.Add(t);
                    this.listBoxStatistika.Items.Add(tocke[i]);
                }
                //stvaranje matrice udaljenosti
                udaljenosti = new double[brGradova][];
                for (int i = 0; i < brGradova; i++)
                    udaljenosti[i] = new double[brGradova];

                for (int i = 0; i < brGradova - 1; i++)
                {
                    PointF gradI = (PointF)tocke[i];
                    for (int j = i + 1; j < brGradova; j++)
                    {
                        PointF gradJ = (PointF)tocke[j];

                        double udaljenost = Math.Sqrt(Math.Pow(gradI.X - gradJ.X, 2) + Math.Pow(gradI.Y - gradJ.Y, 2));
                        udaljenosti[i][j] = udaljenost;
                        udaljenosti[j][i] = udaljenost;
                    }
                    udaljenosti[i][i] = 0;
                }

            }
            this.buttonKreni.Visible = true;
        }

        /// <summary>
        /// Metoda koja rauna dobrotu pojedinog kromosoma. Dobrota je izraena kao duljina puta.
        /// </summary>
        private double izracunajDobrotu(ArrayList kromosom)
        {
            double dobrota = 0;

            foreach (int i in kromosom)
            {
                int next = (kromosom.IndexOf(i) + 1) % brGradova;
                dobrota += udaljenosti[i][(int)kromosom[next]];
            }
            return dobrota;
        }
        /// <summary>
        /// Metoda koja radi PMX krizanje
        /// </summary>
        private Kromosom PartiallyMatchedCrossover(Kromosom Rod1, Kromosom Rod2, int tockaP1, int tockaP2)
        {
            ArrayList elementi1 = new ArrayList();
            ArrayList elementi2 = new ArrayList();
            
            for (int i = 0; i < brGradova; i++)
            {
                elementi1.Add(null);
                elementi2.Add(null);
            }
            //dodavanje dijelova izmedju tocaka prekida
            for (int i = tockaP1 + 1; i <= tockaP2; i++)
            {
                elementi2[i] = Rod1.Chromosome[i];
                elementi1[i] = Rod2.Chromosome[i];
            }
            //generiranje ostatka kromosoma prema definiciji PMX krizanja
            for (int i = 0; i < brGradova; i++)
            {
                if ((i > tockaP1) && (i <= tockaP2))
                    continue;

                if (elementi1.Contains(Rod1.Chromosome[i]))
                {
                    int index = Rod2.Chromosome.IndexOf(Rod1.Chromosome[i]);
                    while (true)
                    {
                        if (elementi1.Contains(Rod1.Chromosome[index]))
                        {
                            index = Rod2.Chromosome.IndexOf(Rod1.Chromosome[index]);
                            //elementi1[i] = Rod1.Chromosome[index];
                        }
                        else
                        {
                            elementi1[i] = Rod1.Chromosome[index];
                            break;
                        }
                    }
                }
                else
                    elementi1[i] = (int)Rod1.Chromosome[i];

                //drugo dijete

                if (elementi2.Contains(Rod2.Chromosome[i]))
                {
                    int index = Rod1.Chromosome.IndexOf(Rod2.Chromosome[i]);
                    while (true)
                    {
                        if (elementi2.Contains(Rod2.Chromosome[index]))
                        {
                            index = Rod1.Chromosome.IndexOf(Rod2.Chromosome[index]);
                            //elementi2[i] = Rod2.Chromosome[index];
                        }
                        else
                        {
                            elementi2[i] = Rod2.Chromosome[index];
                            break;
                        }
                    }
                }
                else
                    elementi2[i] = (int)Rod2.Chromosome[i];
            }
            //stvaranje djece
            Kromosom dijete1 = new Kromosom(elementi1, this.izracunajDobrotu(elementi1));
            Kromosom dijete2 = new Kromosom(elementi2, this.izracunajDobrotu(elementi2));
            
            //vracanje boljeg djeteta
            if (dijete1.DuljinaPuta < dijete2.DuljinaPuta)
                return dijete1;

            return dijete2;
        }

        /// <summary>
        /// Metoda koja radi GSX krizanje prema pseudokodu u dokumentaciji rada
        /// </summary>
        public Kromosom GreedySubtourCrossover(Kromosom Rod1, Kromosom Rod2)
        {
            ArrayList elementi = new ArrayList();
            bool fa = true, fb = true;
            Random rnd = new Random();

            int sluc = rnd.Next(brGradova);
            int indR1 = Rod1.Chromosome.IndexOf(sluc);
            int indR2 = Rod2.Chromosome.IndexOf(sluc);
            elementi.Add(sluc);
            do
            {
                if (indR1 == 0)
                    indR1 = 8;
                indR1 = (indR1 - 1) % brGradova;
                indR2 = (indR2 + 1) % brGradova;
                if (fa == true)
                {
                    if (elementi.Contains(Rod1.Chromosome[indR1]))
                    {
                        fa = false;
                    }
                    else
                    {
                        elementi.Add(null);
                        for (int i = elementi.Count - 1; i > 0; i--)
                        {
                            elementi[i] = elementi[i - 1];
                        }
                        elementi[0] = Rod1.Chromosome[indR1];
                    }
                }
                if (fb == true)
                {
                    if (elementi.Contains(Rod2.Chromosome[indR2]))
                    {
                        fb = false;
                    }
                    else
                    {
                        elementi.Add(Rod2.Chromosome[indR2]);
                    }
                }

            } while ((fa == true) || (fb == true));

            while (elementi.Count < brGradova)
            {
                sluc = rnd.Next(brGradova);
                if (elementi.Contains(sluc))
                    continue;
                else
                    elementi.Add(sluc);
            }

            Kromosom dijete = new Kromosom(elementi, this.izracunajDobrotu(elementi));
            return dijete;
        }


        /// <summary>
        /// Metoda koja radi jednostavnu mutaciju
        /// </summary>
        private Kromosom jednostavnaMutacija(Kromosom zaMutirati, string str_vM)
        {
            bool promili = true;
            int vjer = -1;
            int vM = -1;
            if (str_vM.EndsWith("pp"))
            {
                vM = System.Convert.ToInt32(str_vM.Remove(str_vM.IndexOf("p")));
            }
            else
            {
                vM = System.Convert.ToInt32(str_vM.Remove(str_vM.IndexOf("p")));
                promili = false;
            }

            for (int i = 0; i < zaMutirati.Chromosome.Count; i++)
            {
                if (promili)
                {
                    vjer = rnd.Next(1000);

                    if (vjer <= vM)
                    {
                        int rndGrad = rnd.Next(zaMutirati.Chromosome.Count);
                        zaMutirati.Chromosome[zaMutirati.Chromosome.IndexOf(rndGrad)] = zaMutirati.Chromosome[i];
                        zaMutirati.Chromosome[i] = rndGrad;
                    }

                }
                else
                {
                    vjer = rnd.Next(100);

                    if (vjer <= vM)
                    {
                        int rndGrad = rnd.Next(zaMutirati.Chromosome.Count);
                        zaMutirati.Chromosome[zaMutirati.Chromosome.IndexOf(rndGrad)] = zaMutirati.Chromosome[i];
                        zaMutirati.Chromosome[i] = rndGrad;
                    }

                }
            }
            zaMutirati.DuljinaPuta = izracunajDobrotu(zaMutirati.Chromosome);
            return zaMutirati;
        }

        public Kromosom nadjiNajboljeg()
        {
            Kromosom najbolji = (Kromosom)this.Populacija[0];
            foreach (Kromosom k in this.Populacija)
            {
                if (izracunajDobrotu(najbolji.Chromosome) > izracunajDobrotu(k.Chromosome))
                    najbolji = k;
            }
            return najbolji;
        }
        private void inicijaliziraj(int velPopulacije)
        {
            Populacija.Clear();
            ArrayList elementi = new ArrayList();
            for (int i = 0; i < velPopulacije; i++)
            {
                for (int j = 0; j < brGradova; j++)
                {
                    int rndBroj = rnd.Next(brGradova);
                    if (elementi.Contains(rndBroj))
                    {
                        j--;
                    }
                    else
                    {
                        elementi.Add(rndBroj);

                    }
                }
                //generiranje novog kromosoma iz stvorene staze
                Kromosom jedinka = new Kromosom(elementi, this.izracunajDobrotu(elementi));
                Populacija.Add(jedinka);
                //this.listBoxStatistika.Items.Add(jedinka.ToString());
               
                elementi.Clear();
            }
        }

        private void buttonKreni_Click(object sender, EventArgs e)
        {
            this.algoritam();
        }
    }
}